#
# Predictive Modelling Metadata Interchange Format
#
# Requires OrderedDict (available in collections in Python 2.7 and backported)
#
from __future__ import division
from __future__ import unicode_literals
import json
import datetime
import sys
from collections import OrderedDict, Counter

PMM_VERSION = '0.1'
DEFAULT_DATE_TAG_FORMAT = '%Y-%m-%d %H:%M:%S'

if sys.version_info.major < 3:
    STRING_TYPE = unicode
else:
    STRING_TYPE = str



class PMMError(Exception):
    pass


class TAG(object):
    CATEGORICAL = 'categorical'
    ORDINAL = 'ordinal'
    UNIQUE = 'unique'

    MAXIMIZE = 'maximize'
    MINIMIZE = 'minimize'


class ROLE(object):
    INDEPENDENT = 'independent'         # left-hand side; predictor
    DEPENDENT = 'dependent'             # right-hand side; outcome
    TREATMENT = 'treatment'             # specifies which treatment, if any
    WEIGHT = 'weight'                   # weight field of some kind
    AUXILIARY = 'auxiliary'             # auxiliary field e.g. value field
    VALIDATION = 'validation'           # field specifies cross-validation
                                        #   partition
    IGNORE = 'ignore'                   # ignore other unspecified field
    UNSPECIFIED = ''                    #

    ROLES = (INDEPENDENT, DEPENDENT, TREATMENT, WEIGHT, AUXILIARY,
             VALIDATION, IGNORE, UNSPECIFIED)


FIELD_TYPES = ['boolean', 'integer', 'real', 'string', 'datestamp']


class PMMType(object):
    """
    Template class used to specify each element of the PMM format.

    Typically, you simply need to subclass PMMType and list the required,
    defaulted, and optional arguments. This will provide a constructor with the
    normal Python style (i.e. supply required and defaulted arguments with a
    combination of positional and keyword args, and optional arguments using
    keyword ags).  You also get a serializable() method that turns the
    element into an ordered dictionary (to write out as JSON say), and which
    can normally be passed back to the constructor as keyword arguments to
    reconstruct the instance, e.g. elt2 = MyElt(**elt.serializable())

    In more complex cases you can override the constructor and serializable
    methods.
    """

    required = OrderedDict(())   # name -> type
    defaulted = OrderedDict(())  # name -> (type, default)
    optional = OrderedDict(())   # name -> type
    # note use type=any to avoid specifying a specific type for an attribute

    def __init__(self, *args, **kwargs):
        """
        Generic initializer to set up class with positional and keyword
        arguments corresponding to the required, defaulted and optional args
        defined for the class
        """

        # convenience class attribute with type for each attribute
        if not hasattr(self.__class__, 'typemap'):
            self.__class__.typemap = dict([(name, type_)
                                          for (name, (type_, default_))
                                              in self.defaulted.items()])
            self.__class__.typemap.update(self.required)
            self.__class__.typemap.update(self.optional)

        # first do the keyword arguments
        for key, val in kwargs.items():
            self._setattr(key, val)

        # then do the positional paramters
        keys = list(self.required.keys()) + list(self.defaulted.keys())
        if len(args) > len(keys):
            raise PMMError('Constructor for %s takes at most %d '
                           'positional arguments, %d given'
                           % (self.__class__.__name__, len(keys), len(args)))

        for key, val in zip(keys, args):
            self._setattr(key, val)

        # now do any remaining defaulted args
        for key, (type_, default_) in self.defaulted.items():
            if not hasattr(self, key):
                self._setattr(key, default_)

        # check that we've got all the required args
        for key in self.required:
            if not hasattr(self, key) or self.__dict__[key] is None:
                raise PMMError('Constructor for %s missing required '
                               'argument %s' % (self.__class__.__name__,
                                key))

    def _setattr(self, key, val):
        if key not in self.typemap:
            raise PMMError('Unknown attribute %s for class %s'
                           % (key, self.__class__.__name__))
        elif val is None:
            return

        def convert(val, type_):
            # if type_ is a list of types, process the inside of the list
            if (type(type_) in (tuple, list)
                    and len(type_) == 1
                    and type(val) in (tuple, list)):
                return [convert(v, type_[0]) for v in val]

            # if type_ is unspecified or val already matches, we're good
            elif type_ is any or type(val) == type_:
                return val

            # if it's a PMMType instantiated with a dict, treat as kwargs
            elif issubclass(type_, PMMType) and type(val) == dict:
                return type_(**val)

            elif type_ == str and sys.version_info.major < 3:
                return (val.encode('UTF-8') if type(val) == unicode
                        else type_(val))

            # otherwise let type_ try to construct an instance from val
            # (will raise an error if it's incompatible)
            else:
                return type_(val)

        setattr(self, key, convert(val, self.typemap[key]))

    def serializable(self):
        """
        Return a serializable (typically ordered) dict representing self
        """

        # helper to recursively serialize children
        def serialize(val):
            if hasattr(val, 'serializable'):
                return val.serializable()
            elif type(val) in (list, tuple):
                return [serialize(v) for v in val]
            else:
                return val

        dct = OrderedDict([])
        for key in (list(self.required.keys()) + list(self.defaulted.keys())
                    + list(self.optional.keys())):
            if hasattr(self, key):
                dct[key] = serialize(getattr(self, key))
        return dct


class FlatFileFormat(PMMType):
    defaulted = OrderedDict([
            ('encoding', (str, 'UTF-8')),
            ('separator', (str, ',')),
            ('quote', (str, '"')),
            ('escape', (str, '\\')),
            ('nullmarker', (str, '')),
            ('headerrowcount', (int, 1)),
            ('dateformat', (str, None)),
    ])
    # TODO: Date format


class FlatFile(PMMType):
    required = OrderedDict([
            ('name', str),
            ('format', FlatFileFormat)
    ])


class Data(PMMType):
    # at some point we might allow a choice of representations
    required = OrderedDict([
            ('flatfile', FlatFile)
    ])


class Stats(PMMType):
    optional = OrderedDict([
            ('nnulls', int),
            ('nuniques', int),
            ('min', any),     # abuse the any() function to allow any type
            ('max', any),
            ('mean', float),
    ])


class Field(PMMType):
    required = OrderedDict([
            ('name', str),
            ('type', str),
            ('role', str),
            ('tags', dict),      # dictionary of values
            ('stats', Stats),
    ])
    optional = OrderedDict([
            ('values', [any]),  # list of arbitrary type (field dependent)
            ('longname', str),
            ('description', str),
    ])


class Metadata(PMMType):
    required = OrderedDict([
            ('pmmversion', str),       # omitted in direct constructor
            ('name', str),
            ('recordcount', int),
            ('fieldcount', int),       # omitted in direct constructor
            ('fields', [Field]),       # list of Field ojects
            ('tags', dict),            # dictionary of values
    ])
    optional = OrderedDict([
            ('data', Data),
            ('description', str),
            ('creator', str),
            ('contributor', str),
            ('permissions', str),
            ('datetagformat', str),
    ])

    def __init__(self, *args, **kwargs):
        """
        Instantiate without pmmversion and fieldcount when constructing
        directly, e.g.

            Metadata(name, recordcount, fields, ...)

        """

        if len(args) >= 2:
            # automatically construct the pmmversion and fieldcount args
            args = list(args)
            args.insert(0, PMM_VERSION)
            args.insert(3, len(args[3]))

        super(Metadata, self).__init__(*args, **kwargs)

        if self.fieldcount != len(self.fields):
            raise PMMError('Metadata fieldcount %d <> number of fields %d'
                           % (self.fieldcount, len(self.fields)))

        if float(self.pmmversion) != float(PMM_VERSION):
            raise PMMError("Can't handle pmmversion %s (vs %s)"
                           % (self.pmmversion, PMM_VERSION))


    def toJSON(self):
        self.convert_all_date_tags()
        self.order_all_tags()
        jsonStr = json.dumps(self.serializable(), indent=4)
        self.unconvert_all_date_tags()
        return u'\n'.join(L.rstrip() for L in jsonStr.splitlines())

    def order_all_tags(self):
        self.tags = self.order_tags(getattr(self, 'tags', {}))
        for field in self.fields:
            field.tags = self.order_tags(getattr(field, 'tags', {}))

    def order_tags(self, tags):
        d = OrderedDict()
        for tag in sorted(tags):
            d[tag] = tags[tag]
        return d

    def validate_fields(self):
        """
        Validate that all fields have names and types,
        and that the types are allowed PMM types.
        Also checks that names are unique, though does not place any
        other restrictions on the names.
        """
        c = Counter()
        for i, f in enumerate(self.fields):
            if not hasattr(f, 'name'):
                raise PMMError('Unknown name for field %d' % i)
            if not hasattr(f, 'type'):
                raise PMMError('Missing type for field %s' % f.name)
            if not f.type in FIELD_TYPES:
                raise PMMError('Unknown type %s for field %s' % (f.type,
                                                                 f.name))
            c[f.name] += 1
        if any(v > 1 for v in c.values()):
            raise PMMError('Not all field names are unique:',
                           ' '.join(k for k in c if c[k] > 1))

    def validate(self):
        """
        Validate various aspects of a PMM file.
        """
        self.validate_fields()

    def convert_all_date_tags(self):
        datetagformat = getattr(self, 'datetagformat', DEFAULT_DATE_TAG_FORMAT)
        nDates = self.convert_date_tags(getattr(self, 'tags', OrderedDict()),
                                        datetagformat)
        for f in self.fields:
            nDates += self.convert_date_tags(getattr(f, 'tags', OrderedDict()),
                                             datetagformat)
        if nDates:
            self.datetagformat = datetagformat  # ensure set

    def convert_date_tags(self, tags, fmt):
        n = 0
        for tag in tags:
            val = tags[tag]
            if type(val) == datetime.datetime:
                tags[tag] = val.strftime(fmt)
                n += 1
        return n

    def unconvert_all_date_tags(self):
        datetagformat = getattr(self, 'datetagformat', DEFAULT_DATE_TAG_FORMAT)
        self.unconvert_date_tags(getattr(self, 'tags', OrderedDict()),
                                 datetagformat)
        for f in self.fields:
            self.unconvert_date_tags(getattr(f, 'tags', OrderedDict()),
                                     datetagformat)

    def unconvert_date_tags(self, tags, fmt):
        for tag in tags:
            val = tags[tag]
            if type(val) == STRING_TYPE:
                try:
                    tags[tag] = datetime.datetime.strptime(val, fmt)
                except ValueError:
                    pass

    def save(self, path):
        with open(path, 'w') as f:
            f.write(self.toJSON())

    def __getitem__(self, name):
        for field in self.fields:
            if field.name == name:
                return field
        raise KeyError(name)

    def add_field(self, name, fieldMetadata):
        if getattr(fieldMetadata, 'name', None) is None:
            raise Exception('Attempt to add %s metadata with no name: %s'
                            % (name, str(fieldMetadata)))
        for i, field in enumerate(self.fields):
            if field.name == name:
                self.fields[i] = fieldMetadata
                return
        self.fields.append(fieldMetadata)


def interpret_all_date_tags(m):
    datetagformat = getattr(m, 'datetagformat', None)
    if not datetagformat:
        return
    interpret_date_tags(getattr(m, 'tags', OrderedDict()) , datetagformat)
    for f in getattr(m, 'fields', OrderedDict()):
        interpret_date_tags(getattr(f, 'tags', OrderedDict()), datetagformat)


def interpret_date_tags(tags, datetagformat):
    for tag in tags:
        v = tags[tag]
        if type(v) == STRING_TYPE:
            try:
                dt = datetime.datetime.strptime(v, datetagformat)
                tags[tag] = dt
            except ValueError:
                pass


def load(path):
    """
    Reads a PMM file from the path given.
    Returns a PMM Metadata object.
    """
    with open(path) as f:
        jsonText = f.read()
    return loads(jsonText)


def loads(jsonText):
    """
    Converts the JSON text given into a Metadata object.
    """
    m = Metadata(**json.loads(jsonText))
    interpret_all_date_tags(m)
    return m





# -*- coding: utf-8 -*-
"""
featherpmm.py: Extends feather file format with a paired file
that contains extra metadata. If the feather file is /path/to/foo.feather,
the metadata file is /path/to/foo.pmm.
"""

from __future__ import division

import os
import datetime
import numpy as np
import sys

try:
    import feather
except ImportError:
    feather = None

from pmmif import pmm

PANDAS_STRING_DTYPE = np.dtype('O')
PANDAS_BOOL_DTYPE = np.dtype('bool')
PANDAS_FLOAT_DTYPE = np.dtype('float64')
NULL_STRING = '∅'
NULL_SUFFIX = '_' + NULL_STRING
UTF8 = 'UTF-8'

isPython2 = sys.version_info.major < 3
if isPython2:
    bytes_type = str                # type: type
    unicode_type = unicode          # type: type
else:
    bytes_type = bytes
    unicode_type = str


class Dataset(object):
    """
    Container for a Pandas dataframe and a metadata object from PMM.

    Metadata fulfils two functions:
        1. It allows association of 'intended' types of columns, in cases
           for which Pandas forces promotion. For example:
                - Integer columns that contain nulls are promoted to float
                  in Pandas, but can be marked as 'integer' in the metadata.
                - Boolean columns that contain nulls are promoted to Object
                  in Pandas, but can be marked as 'boolean' in the metadata.
        2. It allows additional annotations to be associated with the
           dataframe, and with individual columns of the dataframe.
                - Tags (with optional values)
                - Descriptions

    A dataset object has two attributes:
          - df    is an ordinary Pandas dataframe
          - md    is a PMM Metadata object
    """
    def __init__(self, df, md=None, name=None):
        """
        Create Dataset object.
          -  df   is a Pandas dataframe.
          -  md   is a PMM metadata object, and is optional.
          -  name is the name to be associated with the dataset, for the
                  case where no metadata is provided.

        If no metadata is provided, the metadata is inferred from the
        dataframe.
        """
        self.df = df
        self.md = md or _create_pmm_metadata(df, name)

    def add_field(self, name, col, pmmtype=None):
        """
        Add a new field to the dataset.

        Adds the field to the Pandas dataframe, and declares its type
        in the PMM metadata.

        The type, if provided, must be one of: 'boolean', 'integer',
        'real', 'string', 'datestamp'. If the type is not specified, the
        type is inferred from the dataframe.

        The advantages of using add_field, as opposed to just creating it
        directly in the dataframe, are:
            - the option of specifying an intended type, even if Pandas has
              promoted the type in the dataframe.
            - creating a corresponding entry in the metadata, to facilitate
              tagging, descriptions, etc.
        """
        self.df[name] = col
        self.declare_field(name, pmmtype)

    def declare_field(self, name, pmmtype=None):
        """
        Declare the type of a field in the dataset, which must already
        exist in the Pandas dataframe.

        This is intended for use when a dataframe has been created without
        any metadata (for example, from a CSV file), and then more detailed
        type information needs to be declared for existing fields.

        The type, if provided, must be one of the types described for
        add_field, above.
        """
        if self.df[name].isnull().all():  # all null or no records
            if pmmtype == 'string':
                self.df[name] = self.df[name].astype(np.dtype('O'))
            elif pmmtype == 'datestamp':
                self.df[name] = (self.df[name]
                                     .astype(np.dtype('datetime64[ns]')))
        fieldMetadata = _create_pmm_field(self.df[name], pmmtype=pmmtype)
        self.md.add_field(name, fieldMetadata)

    def tag_field(self, colname, tagname, value=None):
        """
        Add a tag to a field. The tag can optionally have a value, but
        by default does not.

        The field must already exist in the metadata.

        If a value is provided, it must be of one of the following Python
        types:
            - None
            - bool
            - int
            - float
            - str
            - datetime.datetime
        or it can be a (potentially nested) list or dictionary over these
        types with string keys.
        """
        self.md[colname].tags[tagname] = value

    def tag_dataset(self, tagname, value=None):
        """
        Add a tag to the dataset. The tag can optionally have a value, but
        by default does not. If a value is provided, it must have one of
        the types described in tag_field above.
        """
        self.md.tags[tagname] = value

    def update_metadata(self):
        """
        Update the metadata to bring it into line with the dataset.

        After calling this method, all of the fields that exist in the
        dataset will now exist in the metadata too, and the metadata will
        not contain any fields that do not appear in the dataframe.

        It will infer types in the metadata for any fields that do not
        already have metadata, but will not alter the types of existing
        fields in the metadata.
        """
        _reset_fields_from_dataframe(self)

    def merge_metadata(self, other, fields=None):
        """
        After merging another dataset in (via pd.merge), we don't have any of
        the metadata associated with the fields that have come in from that
        other dataset. This function takes the 'other metadata' and brings it
        into the metadata for the main dataset. The new metadata wins if the
        fields exists in both if:
            1. the fields parameter is left empty, or
            2. the fields parameter is a list but field is not in the list

        """
        newfields = [fx.name for fx in other.md.fields]
        for f in newfields:
            if (fields is None) or (f in fields):
                if f in [fx.name for fx in self.md.fields]:
                    del self.md.fields[[fx.name
                                        for fx in self.md.fields].index(f)]
        _add_metadata_from_other_dataset(self.md, other.md)
        _reset_fields_from_dataframe(self)

    def append(self, other):
        """
        Append another dataset to an existing one.

        The second dataframe is appended to the first one, and the metadata
        for any fields that only exist in the second one is added to the
        first one.
        """
        self.df = self.df.append(other.df, ignore_index=True)
        _add_metadata_from_other_dataset(self.md, other.md)
        _reset_fields_from_dataframe(self)


def read_dataframe(featherpath):
    """
    Similar to feather.read_dataframe except that it also reads the
    corresponding .pmm file, if present, and returns a Dataset
    object rather than a dataframe.

    The Dataset object contains the Pandas dataframe in its df attribute,
    and the metadata in its md attribute.
    """
    if feather is None:
        raise Exception('Feather-format is not available')
    df = feather.read_dataframe(featherpath)
    pmmpath, datasetname = _split_feather_path(featherpath)
    if os.path.exists(pmmpath):
        md = pmm.load(pmmpath)
    else:
        md = _create_pmm_metadata(df, datasetname)
    df = _recover_problematical_all_null_columns(Dataset(df, md))
    return Dataset(df, md)


def write_dataframe(dataset, featherpath):
    """
    Similar to feather.write_dataframe except that it also writes a
    corresponding .pmm file, and expects a Dataset object rather than a
    dataframe.

    The Dataset object contains the Pandas dataframe in its df attribute,
    and the metadata in its md attribute.
    """
    if feather is None:
        raise Exception('Feather-format is not available')
    pmmpath, datasetname = _split_feather_path(featherpath)
    if dataset.md is None:
        dataset.md = _create_pmm_metadata(dataset.df, datasetname)

    _reset_fields_from_dataframe(dataset)

    df = _sanitize_problematical_all_null_columns(dataset)
    try:
        feather.write_dataframe(df, featherpath)
        dataset.md.save(pmmpath)
    except:
        # feather leaves dud files around if it fails to write
        if os.path.exists(featherpath):
            os.remove(featherpath)
        if os.path.exists(pmmpath):
            os.remove(pmmpath)
        raise


#
# The rest of the functions below are internal to this module, and should not
# be called from outside.
#


def _sanitize_problematical_all_null_columns(ds):
    """
    Feather doesn't like all-null string columns or all-null boolean columns,
    so this method transforms them before saving.
    They are transformed into float64 fields with NaN et every value,
    and the pandas column name gets a '_∅t' appended, where t is a
    type indicator --- b for boolean, s for string or u for unknown.
    """
    origdf, md = ds.df, ds.md
    df = origdf[list(origdf)]       # Copye
    nTransformed = 0
    fieldnames = list(df.columns)
    nRecords = len(df.index)
    dfIsUnicode = fieldnames and type(fieldnames[0]) == unicode_type
    fn = _unicode_definite if dfIsUnicode else _utf8_definite
    null_suffix = fn(NULL_SUFFIX)
    for i, f in enumerate(fieldnames):
        if (df[f].dtype not in (PANDAS_STRING_DTYPE, PANDAS_BOOL_DTYPE)
                or df[f].notnull().sum() > 0):  # includes bools with nulls
            continue
        if df[f].dtype == PANDAS_BOOL_DTYPE and nRecords > 0:
            continue
        typeChar = ('b' if md[f].type == 'boolean'
                        else 's' if md[f].type == 'string'
                        else 'u')
        altname = f + null_suffix + typeChar
        if altname in fieldnames:
            continue  # already there. Whatever...
            # restore any all-null string fields
        df[altname] = np.array([np.nan] * nRecords,
                               dtype=PANDAS_FLOAT_DTYPE)
        nTransformed += 1
        fieldnames[i] = altname
    if nTransformed > 0:
        df = df[fieldnames]
    return df


def _recover_problematical_all_null_columns(ds):
    """
    Feather doesn't like all-null string columns or all-null boolean columns,
    so they are sanitized before saving; this untransforms them.
    """
    df, md = ds.df, ds.md
    nTransformed = 0
    fieldnames = list(df.columns)
    nRecords = len(df.index)
    dfIsUnicode = fieldnames and type(fieldnames[0]) == unicode_type
    fn = _unicode_definite if dfIsUnicode else _utf8_definite
    null_suffix = fn(NULL_SUFFIX)
    for i, f in enumerate(fieldnames):
        if f[:-1].endswith(null_suffix) and all(np.isnan(df[f])):
            suffixLen = len(null_suffix) + 1
            suffix = f[-suffixLen:]
            truename = f[:-suffixLen]
            typeChar  = suffix[-1]
            if truename in df:
                continue  # Both there; makes no sense; leave
            # restore any all-null string fields
            type_ = PANDAS_STRING_DTYPE
            if (typeChar == 'b' and nRecords == 0):
                type_ = PANDAS_BOOL_DTYPE
            df[truename] = np.array(np.ones(nRecords) * np.nan, type_)
            nTransformed += 1
            fieldnames[i] = truename
    if nTransformed > 0:
        df = df[fieldnames]
    return df


def _split_feather_path(path):
    """
    Returns path to the corresponding PMM file and the dataset name
    from the feather path given.
    """
    pathbody, ext = os.path.splitext(path)
    if ext.startswith('.feather'):
        pmmpath = pathbody + '.pmm' + ext[8:]
    else:
        pmmpath = pathbody + '.pmm'
    datasetname = os.path.split(pathbody)[1]
    return pmmpath, datasetname


def _create_pmm_metadata(df, name):
    """
    Creates a vanilla PMM Metadata struture from a data frame.
    """
    nRecords = len(df.index)
    fields = [_create_pmm_field(df[col]) for col in df]
    return pmm.Metadata(name, nRecords, fields, tags={})


def _create_pmm_field(col, pmmtype=None):
    """
    Creates a PMM Field object from a column.
    """
    return pmm.Field(col.name, _pmm_type(col, pmmtype), role='',
                     tags={}, stats={})


def _pmm_type(col, pmmtype=None):
    """
    Maps Pandas types to PMM types
    """
    if pmmtype:
        if pmmtype not in pmm.FIELD_TYPES:
            raise ValueError('Unknown PMM type: %s:' % pmmtype)
        return pmmtype
    s = str(col.dtype)
    if s == 'bool':
        return 'boolean'
    elif s.startswith('int'):
        return 'integer'
    elif s.startswith('float'):
        return 'real'
    elif s == 'object':
        return 'string'
    elif s.startswith('date'):
        return 'datestamp'
    else:
        raise TypeError('Unknown type: %s [%s]' % (s, repr(col.dtype)))


def _reset_fields_from_dataframe(dataset):
    """
    Make some effort to ensure that the metadata matches the data
    before writing.
    """
    df, md = dataset.df, dataset.md
    dfFieldnames = list(df)
    mdFieldnames = [f.name for f in md.fields]

    dfOnlyFieldnames = set(dfFieldnames) - set(mdFieldnames)
    mdOnlyFieldnames = set(mdFieldnames) - set(dfFieldnames)

    for f in dfOnlyFieldnames:
        md.fields.append(_create_pmm_field(df[f]))
    assert(mdOnlyFieldnames.intersection(set([f.name for f in md.fields]))
           == mdOnlyFieldnames)

    # Remove metadata for fields that aren't in the dataframe
    for f in mdOnlyFieldnames:
        del md.fields[[fx.name for fx in md.fields].index(f)]

    orderOK = True
    for i, f in enumerate(list(md.fields)):
        if not f.name == dfFieldnames[i]:
            orderOK = False
    if not orderOK:
        md.fields = [md[f] for f in dfFieldnames]
    md.fieldcount = len(md.fields)
    md.recordcount = len(df)


def _add_metadata_from_other_dataset(md, othermd):
    """
    Extend md with fields from othermd not present in md.
    (No consistency checking.)
    """
    fields = [f.name for f in md.fields]
    otherFields = [f.name for f in othermd.fields]
    otherOnlyFields = set(otherFields) - set(fields)
    for f in otherOnlyFields:
        md.fields.append(othermd[f])
    md.fieldcount = len(md.fields)


def _unicode_definite(s):
    """
    Converts a string to unicode if it isn't.
    Assumes encoding is UTF-8.
    Stringifies (as unicode) non strings.
    """
    return unicode_type(s, UTF8) if type(s) == bytes_type else unicode_type(s)


def _utf8_definite(s):
    """
    Converts a string to UTF-8 if it is unicode.
    Otherwise just returns the string.
    Stringifies non strings.
    """
    return s.encode(UTF8) if type(s) == unicode_type else bytes_type(s)


def _utf8_definite_object(s):
    """
    Converts all unicode within scalar or object, recursively, to unicode.
    Handles lists, tuples and dictionaries, as well as scalars.
    """
    if type(s) == unicode_type:
        return s.encode(UTF8)
    elif type(s) == list:
        return [_utf8_definite_object(v) for v in s]
    elif type(s) == tuple:
        return tuple([_utf8_definite_object(v) for v in s])
    elif type(s) == dict:
        return {_utf8_definite_object(k): _utf8_definite_object(v)
                for (k, v) in s.items()}
    return s


def _unicode_definite_object(s):
    """
    Converts all bytes within scalar or object, recursively, to unicode.
    Handles lists, tuples and dictionaries, as well as scalars.
    """
    if type(s) == bytes_type:
        return s.decode(UTF8)
    elif type(s) == list:
        return [_unicode_definite_object(v) for v in s]
    elif type(s) == tuple:
        return tuple([_unicode_definite_object(v) for v in s])
    elif type(s) == dict:
        return {unicode_definite_object(k): _unicode_definite_object(v)
                for (k, v) in s.items()}
    return s


if isPython2:
    _str_definite_object = _utf8_definite_object
    _str_definite = _utf8_definite
else:
    _str_definite_object = _unicode_definite_object
    _str_definite = _unicode_definite



# -*- coding: utf-8 -*-
"""
testpmm.py: tests for pmm.py
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import json
import tempfile
import unittest

from pprint import pprint as pp

from pmmif.pmm import (Metadata, Field, Stats, Data, FlatFile, FlatFileFormat,
                       ROLE, TAG, load)

HILLSTROM_PMM_VERSION = '0.1'


class LoadTests(unittest.TestCase):
    indir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')
    outdir = tempfile.gettempdir()
    def testRoundTrip(self):
        for pmmfile in ['hillstrom.pmm', 'victorlo.pmm']:
            fullpath = os.path.join(self.indir, pmmfile)
            m = load(fullpath)
            dupfile = os.path.join(self.outdir, pmmfile)
            with open(dupfile,'w') as f:
                f.write(m.toJSON())
            with open(fullpath) as f:
                original = f.read()
            with open(dupfile) as f:
                copy = f.read()
            if not copy == original:
                path = os.path.join(self.outdir, 'failed-' + pmmfile)
                with open(path, 'w') as f:
                    f.write(copy)
                    print('Actual failing output written to %s.' % path)
            self.assertEqual(original, copy)
            try:
                os.unlink(dupfile)
            except:
                pass

    def testLoad(self):
        path = os.path.join(self.indir, 'hillstrom.pmm')
        m = load(path)

        # Basic Metadata

        self.assertEqual(set(m.__dict__.keys()),
                         {'name',
                          'recordcount',
                          'fields',
                          'fieldcount',
                          'data',
                          'pmmversion',
                          'tags'})
        self.assertEqual(m.name, 'hillstrom')
        self.assertEqual(m.recordcount, 64000)
        self.assertEqual(m.fieldcount, 12)
        self.assertEqual(m.pmmversion, HILLSTROM_PMM_VERSION)

        # The data section describes a single flat file
        # named hillstrom.csv with the following format

        self.assertEqual(list(m.data.__dict__.keys()), ['flatfile'])
        f = m.data.flatfile
        self.assertEqual(set(f.__dict__.keys()), {'name', 'format'})
        self.assertEqual(f.name, 'hillstrom.csv')
        fmt = f.format
        self.assertEqual(set(fmt.__dict__.keys()),
                         {
                             'encoding',
                             'escape',
                             'headerrowcount',
                             'nullmarker',
                             'quote',
                             'separator',
                         })

        self.assertEqual(fmt.encoding, 'UTF-8')
        self.assertEqual(fmt.escape, '\\')
        self.assertEqual(fmt.headerrowcount, 1)
        self.assertEqual(fmt.nullmarker, '')
        self.assertEqual(fmt.quote, '"')
        self.assertEqual(fmt.separator, ',')

        # There are 12 fields
        fields = m.fields
        self.assertEqual(len(fields), 12)
        self.assertEqual([f.name for f in fields],
                         [
                             'recency',
                             'history_segment',
                             'history',
                             'mens',
                             'womens',
                             'zip_code',
                             'newbie',
                             'channel',
                             'segment',
                             'visit',
                             'conversion',
                             'spend',
                         ])
        self.assertEqual([f.type for f in fields],
                         [
                             'integer',
                             'string',
                             'real',
                             'boolean',
                             'boolean',
                             'string',
                             'boolean',
                             'string',
                             'string',
                             'boolean',
                             'boolean',
                             'real',
                         ])



def buildVictorLo():
    m = Metadata('victorlo', 99999,
        [
            Field('age','real', ROLE.INDEPENDENT, {}, Stats()),
            Field('trade','integer', ROLE.INDEPENDENT, {TAG.CATEGORICAL: None},
                                     Stats()),
            Field('wealth','real', ROLE.INDEPENDENT, {}, Stats()),
            Field('asset','real', ROLE.INDEPENDENT, {}, Stats()),
            Field('homevalue','real', ROLE.INDEPENDENT, {}, Stats()),
            Field('trt_flg','integer', ROLE.TREATMENT, {}, Stats()),
            Field('respond','integer', ROLE.DEPENDENT, {}, Stats()),
            Field('trn_flg','integer', ROLE.VALIDATION, {}, Stats()),
        ],
        data=Data(FlatFile('VictorLoDRA.dat', FlatFileFormat())),
        description="Synthetic dataset for True Lift paper",
        creator="Victor Lo",
        permissions="Public",
        tags={})

    outpath = os.path.join(LoadTests.outdir, 'victorlo.pmm')
    with open(outpath, 'w') as f:
        f.write(m.toJSON())


class ConstructTests(unittest.TestCase):
    def testVictorLo(self):
        buildVictorLo()



if __name__ == '__main__':
    unittest.main()





import os
from setuptools import setup, find_packages

from pmmif.version import version as __version__

def read(fname):
    # read contents of file
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

def data(path, pathitems):
    # build list of additional files to package up from a subdirectory
    names = []
    for relpath in pathitems:
        subpath = path + [relpath]
        dirname = os.path.join(*subpath)
        for name in os.listdir(dirname):
            pathname = os.path.join(relpath, name)
            fullpathname = os.path.join(dirname, name)
            if os.path.isdir(fullpathname):
                names.extend(data(path, [pathname]))
            else:
                names.append(pathname)
    return names

setup(
    name='pmmif',
    version=__version__,
    author='Stochastic Solutions Limited',
    author_email='info@StochasticSolutions.com',
    description='Predictive Modelling Metadata Interchange Format',
    long_description=read('README.md'),
    license='MIT',
    url='http://www.tdda.info',
    download_url='https://github.com/tdda/pmmif',
    keywords='predictive modelling metadata',
    packages=find_packages(),
    package_data={
        'pmmif': ['README.md', 'LICENSE.txt'] + data(['pmmif'],
                                                     ['doc', 'data']),
    },
    zip_safe=False,
    install_requires=['numpy>=1.9'],
)
